#ifndef _SRNode_hh_
#define _SRNode_hh_

#include <string>
#include <vector>
#include <list>

#include "common.hh"
#include "Token.hh"

class SRNode;
class SRUnion;
class Symbol;
class Tree;
class RENode;

class SRUnion
{
private:
  list<SRNode*>* nodes;
  string doString(string indent) const;
public:
  list<SRNode*>* getNodes() { return nodes; }
  
  vector<RENode*>* appendMatches(Tree * tree, RENodeI ni, Symbol * literal, vector<RENode*>* matches = 0) const;
  string toString() const;
  
  SRUnion(list<SRNode*>* nodeSet=0) : nodes(nodeSet) {};
  ~SRUnion() { delete nodes; }
};

class SRNode
{
private:
  Token * token;
  SRUnion * next;
public:
  Token   * getToken() const { return token; }
  SRUnion * getNext()  const { return next; }
  bool isComplete()    const;// { return (((next->getNodes()==0)) && ((token->getPrevious())==0)); }
  
  vector<RENode*>* appendMatches(Tree * tree, RENodeI ni, Symbol * literal, vector<RENode*>* matches = 0) const;
  string toString() const;
  
  SRNode(Token * tok, SRUnion * nxt) : token(tok), next(nxt) {};
  ~SRNode() {};
};



#ifdef _OLD_STUFF_
class SRNode
{
private:
  /** Children */
  unsigned int children;
  
  /** The Token at this position */
  Token * token;
  /** Parent */
  SRNode * parent;
  /** Next token in inputstring */
  SRNodeUnified * next;
  
  virtual SRNode * getParent() const;
  virtual SRNode * getNext() const;
  virtual Token  * getNextToken() const;
  
public:
  virtual bool isRoot() const;
  virtual bool isComplete() const;
  
  virtual Token  * getToken() const;
  virtual Token  * getPreviousToken() const;
  
  virtual SRNode * shift();
  virtual SRNode * reduce(TokenGroup * group);
  
  //  virtual vector<RENode*>* appendMatches(Tree * tree, RENodeI ni, Symbol * literal);
  virtual vector<RENode*>* appendMatches(Tree * tree, RENodeI ni, Symbol * literal, vector<RENode*>* matches = 0);
  
  virtual string toString() const;
  
  SRNode(Token * tok, SRNode * par=0, SRNodeUnified * nxt=0);
  
  virtual ~SRNode();
};

class SRNodeUnified : public SRNode
{
private:
  vector<SRNode*> * nodes;
  
public:
  virtual bool isRoot() const;
  virtual bool isComplete() const;
  
  virtual vector<RENode*>* appendMatches(Tree * tree, RENodeI ni, Symbol * literal, vector<RENode*>* matches = 0);
  
  virtual string toString() const;
  
  SRNodeUnified(Token * tok, vector<SRNode*>* n) : SRNode(tok,0,0), nodes(n) {};
  
  virtual ~SRNodeUnified() { delete nodes; };
  
};
#endif

#endif
