#include "Rule.hh"
#include "Token.hh"
#include "Symbol.hh"
#include "debug.hh"

unsigned int Token::appendMatchesInvoked = 0;
unsigned int Token::cloneInvoked = 0;
unsigned int Token::constructorInvoked = 0;

extern Synonyms * synonyms;

void Token::setPrevious(Token * p)
{
  prev = p;
}
Token * Token::getPrevious() const
{
  return prev;
}

void Token::doSynonyms(bool doit)
  // does nothing atm.
  // Synonyms are handled in tokenizer.cc
{
  syns = new list<Token*>();
}

bool Token::isPunctuation() const
{
  return word[0] == '$';
}

Token* Token::doRewrite(string fu, string fo, string sfu, string sfo, string w, string b) const
{
  return new Token(fu,fo,sfu,sfo,w,b,false);
}

Token* TokenGroup::doRewrite(string fu, string fo, string sfu, string sfo, string w, string b) const
{
  return new TokenGroup(children,fu,fo,sfu,sfo,w,b,false);
}

Token* TokenSet::doRewrite(string fu, string fo, string sfu, string sfo, string w, string b) const
{
  return new Token(fu,fo,sfu,sfo,w,b,false);
}

bool Token::matches(Symbol* sym) const
{
  bool b=
    (sym->funcs->contains(func) || sym->funcs->tags->size()==0 ||
     (sym->funcs->contains("X")) ) &&
    (sym->forms->contains(form) ||  sym->forms->tags->size()==0 ||
     (sym->forms->contains("x")) ) &&
    (sym->sfuncs->contains(sfunc) ||  sym->sfuncs->tags->size()==0 ||
     (sym->sfuncs->contains("Y")) ) &&
    (sym->sforms->contains(sform) ||  sym->sforms->tags->size()==0 ||
     (sym->sforms->contains("y")) ) &&
    (sym->wordforms->contains(word) || sym->wordforms->tags->size()==0 ) &&
    (sym->baseforms->contains(base) || sym->baseforms->tags->size()==0 );
  
  return b;
}

bool TokenSet::matches(Symbol * sym) const
{
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    if((*i)->matches(sym)) return true;
  return false;
}

vector<Token*>* Token::appendMatches(Symbol* sym, vector<Token*>* v=0) const
{
  ++appendMatchesInvoked;
  if(!v) v = new vector<Token*>();
  if(matches(sym)) v->push_back(clone());
  return v;
}

vector<Token*>* TokenSet::appendMatches(Symbol* sym, vector<Token*>* v=0) const
{
  ++appendMatchesInvoked;
  if(!v) v = new vector<Token*>();
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    (*i)->appendMatches(sym,v);
  return v;
}


bool TokenGroup::hasImpossibleDiscontinuities() const
{
  // list of left-discontinuities to search for
  list<Symbol*>* disclist = new list<Symbol*>();
  // whether a left-discontinous token was found without the matching right-discontinous part.
  bool unmatchedLeftDiscontinuity = false;
  
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; ++i)
    {
      Token * t = *i;
      
      if(t->func[0]=='-')
	{
	  // the token left-discontinous.
	  // test whether it matches any right-discontinous tokens.
	  bool match = false;
	  for(list<Symbol*>::iterator li=disclist->begin(), lz=disclist->end();
	      li!=lz; ++li)
	    {
	      Symbol * sym = *li;
	      if(t->matches(sym))
		{
		  list<Symbol*>::iterator ln = li;
		  ++ln;
		  disclist->erase(li);
		  li = ln;
		  delete sym;
		  match = true;
		}
	    }
	  if(!match) unmatchedLeftDiscontinuity = true;
	}
      if(unmatchedLeftDiscontinuity) break;
      
      if(t->form[t->form.length()-1] == '-')
	{
	  // the token is right-discontinous
	  // add a symbol that will match the left-discontinous part of the discontinous token.
	  Symbol * sym = 
	    t->func[0]=='-' 
	    ? new Symbol(new TagField(t->func), 
			 new TagField(t->form.substr(0,t->form.length()-1))) // this is for tokens that are both left- and right-discontinous.
	    : new Symbol(new TagField("-"+t->func), 
			 new TagField(t->form.substr(0,t->form.length()-1))); // for right-discontinous tokens.
	  
	  disclist->push_back(sym);
	}
    }
  
  bool result = disclist->size()!=0 || unmatchedLeftDiscontinuity;
  
  for(list<Symbol*>::iterator li=disclist->begin(), lz=disclist->end();
      li!=lz; ++li)
    delete *li;
  
  delete disclist;
  
  return result;
}

int TokenGroup::getLeafCount() const
{
  int leaves = 0;
  for(vector<Token*>::iterator 
	i=children->begin(), 
	z=children->end(); 
      i!=z; ++i)
    {
      Token * t = *i;
      leaves += t->getLeafCount();
    }
  return leaves;
}


//
// cloning and stringifying :
//
// BUG: clone() is deprecated. Use of copy constructors is The Right Way (TM). Luckily, the newest algorithm doesn't clone :)



Token * Token::clone() const
{
  ++cloneInvoked;
  Token * t = new Token(func,form,sfunc,sform,word,base);
  return t;
}
string Token::toVISL(int level=0, bool useClosedCorpusFormat = false) const
{
  string s;
  for(int i=0; i<(useClosedCorpusFormat?level-1:level); i++) s += "=";
  s += func + //(sfunc!=""?"#":"") + sfunc +
    ":" + form; // +(sform!=""?"#":"") + sform + 
  if(word!="")
    {
      if(useClosedCorpusFormat)
	{
	  s += "\t" + word;
	}
      else
	{
	  s += "\n";
	  for(int i=0; i<level+1; i++) s += "=";
	  s += word;
	}
    }
  s += "\n";
  return s;
}
string Token::toString() const
{
  return
    ( 
     (func)  + (form!="" ? ":" : "") + 
     (form)  + ((sfunc!="" || sform !="")  ? "_" : "" ) +
     (sfunc) + (sform!="" ? ":" : "" ) +
     (sform) + (base!="" ? "\""+base+"\"" : "")
     + (word!="" ? "\'"+word+"\'" : "")
     );
}

Token * TokenGroup::clone() const
{
  ++cloneInvoked;
  vector<Token*> * v = new vector<Token*>(*children);
  TokenGroup * t = new TokenGroup(v,func,form,sfunc,sform,word,base);
  return t;
}
string TokenGroup::toVISL(int level=0, bool useClosedCorpusFormat = false) const
{
  string s;
  for(int i=0; i<(useClosedCorpusFormat?level-1:level); i++) s += "=";
  s += func + //(sfunc!=""?"#":"") + sfunc +
    ":" + form + //(sform!=""?"#":"") + sform + 
    "\n";
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    s += (*i)->toVISL(level+1, useClosedCorpusFormat);
  return s;
}

string TokenGroup::toString() const
{
  string s =
    ( 
     (func)  + (form!="" ? ":" : "") + 
     (form)  + ((sfunc!="" || sform !="")  ? "_" : "" ) +
     (sfunc) + (sform!="" ? ":" : "" ) +
     (sform) + (word!="" ? "\""+word+"\"" : "")
     );
  s += "( ";
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    s += (*i)->toString() + " ";
  s += ") ";
  return s;
}

string TokenSet::toVISL(int level=0, bool useClosedCorpusFormat = false) const
{
  string s, w, fu, fo;
  for(int i=0; i<(useClosedCorpusFormat?level-1:level); i++) s += "=";
  if(children->size() == 0) return s+"\n";
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    {
      Token * t = *i;
      if(w!="")
	{
	  fo+="|";
	  fu+="|";
	}
      fo += t->form;
      fu += t->func;
      if(w=="") w = t->word;
    }
  s += fu + ":" + fo;
  if(w!="")
    {
      if(useClosedCorpusFormat)
	{
	  s += "\t" + w;
	}
      else
	{
	  s += "\n";
	  for(int i=0; i<level+1; i++) s += "=";
	  s += w;
	}
    }
  s += "\n";
  return s;
}

string TokenSet::toString() const
{
  string s;
  s += "{";
  int n=0;
  for(vector<Token*>::iterator i=children->begin(), z=children->end(); 
      i!=z; i++) 
    s+=(n++?" | ":"")+(*i)->toString();
  s += "} ";
  return s;
}

